# -*- coding: utf-8 -*-
#
# Author: Ingelrest François (Athropos@gmail.com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Library General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301 USA

import gtk, gui, media, modules, os.path, tools, urllib

from tools           import consts
from gettext         import gettext as _
from gobject         import TYPE_STRING, TYPE_INT
from media.track     import Track
from gui.extListview import ExtListView

MOD_INFO = ('Tracklist', 'Tracklist', '', [], True, False)


(
    ROW_ICO,   # Icon drawn in front of the row
    ROW_NUM,   # Track number
    ROW_TIT,   # Track title
    ROW_ART,   # Track artist
    ROW_ALB,   # Track album
    ROW_LEN,   # Track length in seconds
    ROW_PTH    # Full path to the file
) = range(7)


class Tracklist(modules.Module):
    """ This module manages the tracklist """

    def __init__(self):
        """ Constructor """
        modules.Module.__init__(self, (consts.MSG_CMD_TRACKLIST_SET,     consts.MSG_CMD_TRACKLIST_CLR,     consts.MSG_CMD_TRACKLIST_ADD,
                                       consts.MSG_CMD_TOGGLE_PAUSE,      consts.MSG_CMD_NEXT,              consts.MSG_CMD_PREVIOUS,
                                       consts.MSG_EVT_NEED_BUFFER,       consts.MSG_EVT_STOPPED,           consts.MSG_EVT_PAUSED,
                                       consts.MSG_EVT_UNPAUSED,          consts.MSG_EVT_TRACK_ENDED,       consts.MSG_CMD_TRACKLIST_SET_EXT,
                                       consts.MSG_CMD_TRACKLIST_ADD_EXT, consts.MSG_CMD_TRACKLIST_SHUFFLE, consts.MSG_EVT_APP_STARTED))


    def onAppStarted(self):
        """ This is the real initialization function, called when the module has been loaded """
        wTree         = tools.prefs.getWidgetsTree()
        self.popup    = None
        self.playtime = 0
        # Retrieve widgets
        self.window     = wTree.get_widget('win-main')
        self.btnClear   = wTree.get_widget('btn-tracklistClear')
        self.btnRepeat  = wTree.get_widget('btn-tracklistRepeat')
        self.btnShuffle = wTree.get_widget('btn-tracklistShuffle')
        self.btnClear.set_sensitive(False)
        self.btnRepeat.set_sensitive(False)
        self.btnShuffle.set_sensitive(False)
        # Create the list and its columns
        txtLRdr   = gtk.CellRendererText()
        txtRRdr   = gtk.CellRendererText()
        pixbufRdr = gtk.CellRendererPixbuf()
        txtRRdr.set_property('xalign', 1.0)

        columns = (('#',         [(pixbufRdr, gtk.gdk.Pixbuf), (txtRRdr, TYPE_INT)], (ROW_NUM, ROW_TIT),                   False),
                   (_('Title'),  [(txtLRdr, TYPE_STRING)],                           (ROW_TIT,),                           True),
                   (_('Artist'), [(txtLRdr, TYPE_STRING)],                           (ROW_ART, ROW_ALB, ROW_NUM, ROW_TIT), True),
                   (_('Album'),  [(txtLRdr, TYPE_STRING)],                           (ROW_ALB, ROW_NUM, ROW_TIT),          True),
                   (_('Length'), [(txtRRdr, TYPE_INT)],                              (ROW_LEN,),                           False),
                   (None,        [(None, TYPE_STRING)],                              (None,),                              False))

        self.list = ExtListView(columns, True, consts.DND_TARGETS.values())
        self.list.get_column(4).set_cell_data_func(txtRRdr, self.fmtLength)
        self.list.enableDNDReordering()
        wTree.get_widget('scrolledwindow1').add(self.list)
        # GTK handlers
        self.list.connect('extlistview-dnd',            self.onDND)
        self.list.connect('key-press-event',            self.onKeyboard)
        self.list.connect('extlistview-modified',       self.onListModified)
        self.list.connect('extlistview-button-pressed', self.onButtonPressed)

        self.btnClear.connect('clicked',   lambda widget: modules.postMsg(consts.MSG_CMD_TRACKLIST_CLR))
        self.btnRepeat.connect('toggled',  self.onButtonRepeat)
        self.btnShuffle.connect('clicked', lambda widget: modules.postMsg(consts.MSG_CMD_TRACKLIST_SHUFFLE))
        # Try to associate an icon to the shuffle button
        if consts.fileImgShuffle is not None: wTree.get_widget('img-shuffle').set_from_file(consts.fileImgShuffle)
        else:                                 wTree.get_widget('img-shuffle').hide()
        # Try to associate an icon to the repeat button
        if consts.fileImgRepeat is not None: wTree.get_widget('img-repeat').set_from_file(consts.fileImgRepeat)
        else:                                wTree.get_widget('img-repeat').hide()


    # Helper functions
    def hasNext(self):                      return self.list.hasMarkUnder(len(self.list)-1)
    def hasPrev(self):                      return self.list.hasMarkAbove(0)
    def getAllFiles(self):                  return [row[ROW_PTH] for row in self.list.getAllRows()]
    def getAllTracks(self):                 return [Track(row[1:], i, len(self.list)) for i, row in enumerate(self.list.getAllRows())]
    def fmtLength(self, col, cll, mdl, it): cll.set_property('text', tools.sec2str(mdl.get_value(it, ROW_LEN)))


    def __updateButtons(self, sensitive):
        """ Enable/disable bottom buttons based on the sensitive parameter """
        self.btnClear.set_sensitive(sensitive)
        self.btnRepeat.set_sensitive(sensitive)
        self.btnShuffle.set_sensitive(sensitive)


    def jumpTo(self, trackIndex):
        """ Jump to the track located at the given index """
        if self.list.hasMark():
            self.list.setItem(self.list.getMark(), ROW_ICO, consts.icoNull)
        self.list.setMark(trackIndex)
        self.list.scroll_to_cell(trackIndex)
        self.list.setItem(trackIndex, ROW_ICO, consts.icoPlay)
        modules.postMsg(consts.MSG_CMD_PLAY,        {'filename': self.list.getItem(trackIndex, ROW_PTH)})
        modules.postMsg(consts.MSG_EVT_NEW_TRACK,   {'track': Track(self.list.getRow(trackIndex)[1:], trackIndex+1, len(self.list))})
        modules.postMsg(consts.MSG_EVT_TRACK_MOVED, {'hasPrevious': self.hasPrev(), 'hasNext': self.hasNext() or self.btnRepeat.get_active()})


    def onTrackEnded(self):
        """ The current track has ended, jump to the next one if any """
        if   self.hasNext():              self.jumpTo(self.list.getMark()+1)
        elif self.btnRepeat.get_active(): self.jumpTo(0)
        else:                             modules.postMsg(consts.MSG_CMD_STOP)


    def onBufferingNeeded(self):
        """ The current track is close to its end, so we try to buffer the next one to avoid gaps """
        if self.hasNext():                modules.postMsg(consts.MSG_CMD_BUFFER, {'filename': self.list.getItem(self.list.getMark()+1, ROW_PTH)})
        elif self.btnRepeat.get_active(): modules.postMsg(consts.MSG_CMD_BUFFER, {'filename': self.list.getItem(0, ROW_PTH)})


    def insert(self, tracks, readTags=True, position=None):
        """
            Insert some tracks in the tracklist, append them if position is None
            Refer to set() for the meaning of readTags
        """
        if not readTags:
            tracks = [[consts.icoNull] + track[0:6] for track in tracks]                    # Remove "album artist", we don't use it
        else:
            files  = [track[7:] for track in tracks if track.startswith('file://')]
            tracks = [[consts.icoNull] + track[0:6] for track in media.getTracks(files)]    # Remove "album artist", we don't use it

            for track in tracks:
                track[ROW_PTH] = 'file://' + track[ROW_PTH]

        if len(tracks) != 0:
            for track in tracks:
                self.playtime += track[ROW_LEN]

            self.__updateButtons(True)
            self.list.insertRows(tracks, position)


    def set(self, tracks, playNow, readTags=True):
        """
            Replace the tracklist, clear it if tracks is None
            If readTags is True, tracks is just a list of filenames and the tags must be read from them
            If readTags is False, tracks already contains all tags information (e.g., as returned by media.readInfo())
        """
        self.playtime = 0
        self.list.clear()

        if tracks is not None:
            self.insert(tracks, readTags)

        if len(self.list) == 0:
            self.__updateButtons(False)
            modules.postMsg(consts.MSG_CMD_STOP)
            modules.postMsg(consts.MSG_EVT_NEW_TRACKLIST, {'tracks': [], 'playtime': 0})
        elif playNow:
            self.jumpTo(0)


    def onStopped(self):
        """ Playback has been stopped """
        if self.list.hasMark():
            self.list.setItem(self.list.getMark(), ROW_ICO, consts.icoNull)
            self.list.clearMark()


    def onPausedToggled(self, icon):
        """ Switch between paused and unpaused """
        if self.list.hasMark():
            self.list.setItem(self.list.getMark(), ROW_ICO, icon)


    def savePlaylist(self):
        """ Save the current tracklist to a playlist """
        destDir = os.path.dirname(self.list.getItem(0, ROW_PTH))

        if destDir.startswith('file://'): outFile = gui.fileChooser.save(self.window, _('Save playlist'), 'playlist.m3u', destDir[7:])
        else:                             outFile = gui.fileChooser.save(self.window, _('Save playlist'), 'playlist.m3u')

        if outFile is not None:
            allFiles = [file[7:] for file in self.getAllFiles() if file.startswith('file://')]
            media.playlist.save(allFiles, outFile)


    def cropSelection(self):
        """ Remove all tracks but the selection """
        self.playtime = sum([row[ROW_LEN] for row in self.list.iterSelectedRows()])
        self.list.cropSelectedRows()


    def removeSelection(self):
        """ Remove all the currently selected tracks """
        for row in self.list.iterSelectedRows():
            self.playtime -= row[ROW_LEN]
            if row[ROW_ICO] != consts.icoNull:
                modules.postMsg(consts.MSG_CMD_STOP)

        self.list.removeSelectedRows()


    def shuffleTracklist(self):
        """ Shuffle the tracks and ensure that the current track stays visible """
        self.list.shuffle()
        if self.list.hasMark():
            self.list.scroll_to_cell(self.list.getMark())


    def showPopupMenu(self, list, path, button, time):
        """ The index parameter may be None """
        if self.popup is None:
            self.popup = tools.loadGladeFile('TracklistMenu.glade')
            # Connect signals
            self.popup.get_widget('item-save').connect('activate', lambda item: self.savePlaylist())
            self.popup.get_widget('item-crop').connect('activate', lambda item: self.cropSelection())
            self.popup.get_widget('item-remove').connect('activate', lambda item: self.removeSelection())
            self.popup.get_widget('menu-popup').show_all()

        # Enable only relevant menu entries
        self.popup.get_widget('item-remove').set_sensitive(path is not None)
        self.popup.get_widget('item-crop').set_sensitive(path is not None)
        self.popup.get_widget('item-save').set_sensitive(len(list) != 0)
        # Show the popup menu
        self.popup.get_widget('menu-popup').popup(None, None, None, button, time)


    # --== Message handler ==--


    def handleMsg(self, msg, params):
        """ A message has been received """
        if   msg == consts.MSG_EVT_PAUSED:                                   self.onPausedToggled(consts.icoPause)
        elif msg == consts.MSG_EVT_STOPPED:                                  self.onStopped()
        elif msg == consts.MSG_EVT_UNPAUSED:                                 self.onPausedToggled(consts.icoPlay)
        elif msg == consts.MSG_EVT_APP_STARTED:                              self.onAppStarted()
        elif msg == consts.MSG_EVT_TRACK_ENDED:                              self.onTrackEnded()
        elif msg == consts.MSG_EVT_NEED_BUFFER:                              self.onBufferingNeeded()
        elif msg == consts.MSG_CMD_TRACKLIST_CLR:                            self.set(None, False)
        elif msg == consts.MSG_CMD_TRACKLIST_SET:                            self.set(params['files'], params['playNow'])
        elif msg == consts.MSG_CMD_TRACKLIST_ADD:                            self.insert(params['files'])
        elif msg == consts.MSG_CMD_TRACKLIST_ADD_EXT:                        self.insert(params['files'], False)
        elif msg == consts.MSG_CMD_TRACKLIST_SET_EXT:                        self.set(params['files'], params['playNow'], False)
        elif msg == consts.MSG_CMD_TRACKLIST_SHUFFLE:                        self.shuffleTracklist()
        elif msg == consts.MSG_CMD_PREVIOUS and self.hasPrev():              self.jumpTo(self.list.getMark()-1)
        elif msg == consts.MSG_CMD_TOGGLE_PAUSE and not self.list.hasMark(): self.jumpTo(0)
        elif msg == consts.MSG_CMD_NEXT:
            if self.hasNext():                self.jumpTo(self.list.getMark()+1)
            elif self.btnRepeat.get_active(): self.jumpTo(0)


    # --== GTK handlers ==--


    def onButtonRepeat(self, btn):
        """ The 'repeat' button has been pressed """
        if self.list.hasMark():
            modules.postMsg(consts.MSG_EVT_TRACK_MOVED, {'hasPrevious': self.hasPrev(), 'hasNext': self.hasNext() or self.btnRepeat.get_active()})


    def onButtonPressed(self, list, event, path):
        """ Play the selected track on double click, or show a popup menu on right click """
        if event.button == 1 and event.type == gtk.gdk._2BUTTON_PRESS and path is not None:
            self.jumpTo(path[0])
        elif event.button == 3:
            self.showPopupMenu(list, path, event.button, event.time)


    def onKeyboard(self, list, event):
        """ Remove the selection if possible """
        if gtk.gdk.keyval_name(event.keyval) == 'Delete':
            self.removeSelection()


    def onListModified(self, list):
        """ Some rows have been added/removed/moved """
        self.__updateButtons(len(list) != 0)
        modules.postMsg(consts.MSG_EVT_NEW_TRACKLIST, {'tracks': self.getAllTracks(), 'playtime': self.playtime})
        modules.postMsg(consts.MSG_EVT_TRACK_MOVED,   {'hasPrevious': self.hasPrev(), 'hasNext':  self.hasNext()})


    def onDND(self, list, context, x, y, dragData, dndId, time):
        """ External Drag'n'Drop """
        if dragData.data == '':
            context.finish(False, False, time)
            return

        dropInfo = list.get_dest_row_at_pos(x, y)

        # A list of filenames
        if dndId == consts.DND_DAP_URI or dndId == consts.DND_URI:
            tracks   = [urllib.url2pathname(uri) for uri in dragData.data.split()]
            readTags = True
        # A list of tracks with all tags provided
        elif dndId == consts.DND_DAP_TRACKS:
            tracks   = [[urllib.url2pathname(field) for field in track.split()] for track in dragData.data.split('\n')]
            readTags = False

            for track in tracks:
                track[media.NFO_NUM] = int(track[media.NFO_NUM])
                track[media.NFO_LEN] = int(track[media.NFO_LEN])

        if dropInfo is None: self.insert(tracks, readTags)
        else:                self.insert(tracks, readTags, dropInfo[0][0])

        context.finish(True, False, time)
